/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import java.io.IOException;
import java.io.InputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;

import org.eiichiro.jazzmaster.deployment.Deployment;
import org.eiichiro.jazzmaster.inject.Binding;
import org.eiichiro.jazzmaster.scope.Scope;
import org.eiichiro.reverb.lang.ClassResolver;
import org.eiichiro.reverb.lang.ClassResolver.Matcher;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Built-in implementation of {@link Scanner} that scans components from runtime 
 * classpath.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class ClasspathScanner implements Scanner {

	private static Logger logger = LoggerFactory.getLogger(ClasspathScanner.class);
	
	/**
	 * Scans components from runtime classpath.
	 * {@code ClasspathScanner} detects the component by the following 
	 * conditions: 
	 * <ul>
	 * <li>Subclass of {@link Component}</li>
	 * <li>Qualified by {@link Stereotype}</li>
	 * <li>Qualified by {@link Deployment}</li>
	 * <li>Qualified by {@link Binding}</li>
	 * <li>Qualified by {@link Scope}</li>
	 * </ul>
	 * If any exception has occurred in scanning, {@code ClasspathScanner} 
	 * returns empty set.
	 * 
	 * @see org.eiichiro.jazzmaster.Scanner#scan()
	 */
	public Set<Class<?>> scan() {
		try {
			ClassResolver<CtClass> resolver = new CtClassClassResolver();
			logger.info("Scanning components from runtime classpath [" + resolver.urls() + "]");
			Set<CtClass> ctClasses = resolver.resolve(new Matcher<CtClass>() {
				
				public boolean matches(CtClass ctClass) {
					try {
						int modifiers = ctClass.getModifiers();
						
						if (ctClass.isInterface() || Modifier.isAbstract(modifiers)
								|| !Modifier.isPublic(modifiers)
								|| ctClass.hasAnnotation(Builtin.class)) {
							ctClass.detach();
							return false;
						}
						
						CtClass superclass = ctClass.getSuperclass();
						
						while (superclass != null) {
							if (superclass.getName().equals(Component.class.getName())) {
								return true;
							}
							
							superclass = superclass.getSuperclass();
						}
						
						for (Object object : ctClass.getAnnotations()) {
							Class<? extends Annotation> annotationType = ((Annotation) object).annotationType();
							
							if (annotationType.isAnnotationPresent(Stereotype.class) 
									|| annotationType.isAnnotationPresent(Deployment.class) 
									|| annotationType.isAnnotationPresent(Binding.class) 
									|| annotationType.isAnnotationPresent(Scope.class)) {
								return true;
							}
						}
						
					} catch (Exception e) {
//						e.printStackTrace();
					}
					
					ctClass.detach();
					return false;
				}
				
			});
			Set<Class<?>> classes = new HashSet<Class<?>>();
			
			for (CtClass ctClass : ctClasses) {
				try {
					classes.add(Class.forName(ctClass.getName(), true, Thread.currentThread().getContextClassLoader()));
				} catch (Exception e) {
					logger.error("Failed to scan component", e);
				}
			}
			
			logger.info("Components [" + classes.size() + "] scaned");
			return classes;
		} catch (Exception e) {
			logger.error("Failed to scan components: " + e.getMessage(), e);
			return Collections.emptySet();
		}
	}
	
	private class CtClassClassResolver extends ClassResolver<CtClass> {
		
		private ClassPool pool;
		
		private CtClassClassResolver() throws NotFoundException {
			ClassPool pool = ClassPool.getDefault();
			
			for (URL url : urls()) {
				pool.appendClassPath(url.getPath());
			}
			
			this.pool = pool;
		}
		
		@Override
		protected CtClass load(String clazz, InputStream stream) {
			try {
				return pool.makeClass(stream);
			} catch (Exception e) {
				return null;
			}
		}

		@Override
		public Set<CtClass> resolveByAnnotation(final Class<? extends Annotation> annotation) 
				throws IOException {
			Matcher<CtClass> matcher = new Matcher<CtClass>() {
				
				@Override
				public boolean matches(CtClass ctClass) {
					try {
						for (Object object : ctClass.getAnnotations()) {
							Annotation a = (Annotation) object;
							
							if (a.annotationType().equals(annotation)) {
								return true;
							}
						}
						
					} catch (Exception e) {}
					
					ctClass.detach();
					return false;
				}
				
			};
			return resolve(matcher);
		}
		
		@Override
		public Set<CtClass> resolveByInterface(final Class<?> interfaceClass)
				throws IOException {
			Matcher<CtClass> matcher = new Matcher<CtClass>() {

				@Override
				public boolean matches(CtClass ctClass) {
					try {
						for (CtClass c : ctClass.getInterfaces()) {
							if (c.getName().equals(interfaceClass.getName())) {
								return true;
							}
						}
						
						CtClass superclass = ctClass.getSuperclass();
						
						if (superclass != null) {
							return matches(superclass);
						}
						
					} catch (Exception e) {}
					
					ctClass.detach();
					return false;
				}
				
			};
			return resolve(matcher);
		}

		@Override
		public Set<CtClass> resolveByName(final String name) throws IOException {
			Matcher<CtClass> matcher = new Matcher<CtClass>() {
				
				@Override
				public boolean matches(CtClass ctClass) {
					if (ctClass.getName().contains(name)) {
						return true;
					}
					
					ctClass.detach();
					return false;
				}
				
			};
			return resolve(matcher);
		}

		@Override
		public Set<CtClass> resolveBySuperclass(Class<?> superclass)
				throws IOException {
			Matcher<CtClass> matcher = new Matcher<CtClass>() {

				@Override
				public boolean matches(CtClass ctClass) {
					try {
						CtClass superclass = ctClass.getSuperclass();
						
						if (superclass != null) {
							if (superclass.getName().equals(ctClass.getName())) {
								return true;
							} else {
								return matches(superclass);
							}
						}
						
					} catch (Exception e) {
					}
					
					ctClass.detach();
					return false;
				}
				
			};
			return resolve(matcher);
		}
		
	}

}
