/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import static org.eiichiro.jazzmaster.Jazzmaster.*;

import java.lang.annotation.Annotation;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import org.apache.commons.lang.ClassUtils;
import org.eiichiro.jazzmaster.inject.Default;
import org.eiichiro.jazzmaster.inject.Key;
import org.eiichiro.jazzmaster.inject.Name;
import org.eiichiro.jazzmaster.lifecycle.Activated;
import org.eiichiro.jazzmaster.lifecycle.Event;
import org.eiichiro.jazzmaster.scope.Application;
import org.eiichiro.jazzmaster.scope.Context;
import org.eiichiro.jazzmaster.scope.Prototype;
import org.eiichiro.jazzmaster.scope.Request;
import org.eiichiro.jazzmaster.scope.Scope;
import org.eiichiro.jazzmaster.scope.Session;
import org.eiichiro.jazzmaster.scope.Singleton;
import org.eiichiro.jazzmaster.scope.SingletonContext;
import org.eiichiro.jazzmaster.scope.Thread;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Function;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Collections2;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.MapConstraint;
import com.google.common.collect.MapConstraints;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;

/**
 * {@code Container} is the core of Jazzmaster. {@code Container} deploys and 
 * manages Jazzmaster components.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Builtin
@Singleton
@Name("org.eiichiro.jazzmaster.Container")
public class Container {

	private Logger logger = LoggerFactory.getLogger(getClass());
	
	private WeakHashMap<Object, Description<?>> descriptions = new WeakHashMap<Object, Description<?>>();
	
	private Set<Class<?>> installed = new HashSet<Class<?>>();
	
	private ListMultimap<Class<?>, Description<?>> components = ArrayListMultimap.create();
	
	private ListMultimap<Annotation, Description<?>> interceptors = ArrayListMultimap.create();
	
	private Map<Class<? extends Annotation>, Description<?>> contexts 
			= MapConstraints.constrainedMap(
					new HashMap<Class<? extends Annotation>, Description<?>>(), 
					new MapConstraint<Class<? extends Annotation>, Description<?>>() {

		public void checkKeyValue(Class<? extends Annotation> scope,
				Description<?> description) {
			logger.info("Context [" + scope.getSimpleName() + " (" 
					+ description.type().getName() + ")] started");
		}
		
	});
	
	private SingletonContext singleton;
	
	/**
	 * Initializes {@code Container} instance.
	 * Sets up built-in {@link Singleton}, {@link Application}, {@link Thread}, 
	 * {@link Session}, {@link Request} and {@link Prototype} contexts and makes 
	 * {@code Container} instance available in application. This method is 
	 * invoked by {@link Jazzmaster} in bootstrap.
	 */
	@SuppressWarnings("unchecked")
	public void initialize() {
		Class<? extends Context> context = Singleton.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Singleton.class, components.get(context).get(0));
		SingletonContext singleton = new SingletonContext();
		singleton.put((Description<SingletonContext>) contexts.get(Singleton.class), singleton);
		descriptions.put(singleton, contexts.get(Singleton.class));
		this.singleton = singleton;
		install(Container.class);
		Description<Container> container = (Description<Container>) components.get(Container.class).get(0);
		singleton.put(container, this);
		descriptions.put(this, container);
		context = Application.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Application.class, components.get(context).get(0));
		context = Thread.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Thread.class, components.get(context).get(0));
		context = Session.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Session.class, components.get(context).get(0));
		context = Request.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Request.class, components.get(context).get(0));
		context = Prototype.class.getAnnotation(Scope.class).value();
		install(context);
		contexts.put(Prototype.class, components.get(context).get(0));
		logger.info("Container started");
	}
	
	/**
	 * Returns assembled component instance of the specified component type.
	 * If the component is not found or duplicated, this method returns 
	 * <code>null</code>.
	 * 
	 * @param <T> The component type.
	 * @param component The component type.
	 * @return The component instance of the specified component type.
	 */
	public <T> T component(Class<T> component) {
		logger.debug("Component is requested: Type [" + component + "]");
		Preconditions.checkArgument(component != null, 
				"Parameter 'component' must not be [" + component + "]");
		List<Description<?>> descriptions = components.get(component);
		
		if (descriptions.isEmpty()) {
			logger.warn("No component has been installed: Component type ["
					+ component + "]; You must install component on ahead "
					+ "invoking [public <T> void install(Class<T> component)]");
			return null;
		} else if (descriptions.size() > 1) {
			logger.warn("Component is duplicated: Component type [" + component
					+ "]; You must specify more concrete type");
			return null;
		}
		
		return component(component(descriptions.get(0)));
	}
	
	/**
	 * Returns assembled component instance corresponding to the specified key 
	 * (Component type and binding annotations).
	 * <b>NOTE: This method is designed for an internal use.</b>
	 * If the component is not found or duplicated, this method returns 
	 * <code>null</code>. If a {@link Default} qualified component is included 
	 * in the candidate, this method returns it in preference.
	 * 
	 * @param <T> The component type.
	 * @param key The {@link Key} constructed from the injective field to get 
	 * the dependent component.
	 * @return The component instance corresponding to the specified key.
	 */
	public <T> T component(final Key<T> key) {
		logger.debug("Component is requested: Key [" + key + "]");
		Collection<Description<?>> descriptions = components.get(key.type());
		
		if (descriptions.isEmpty()) {
			logger.warn("No component has been installed: Component key ["
					+ key + "]; You must install component on ahead "
					+ "invoking [public <T> void install(Class<T> component)]");
			return null;
		}
		
		descriptions = Collections2.filter(descriptions, new Predicate<Description<?>>() {

			public boolean apply(Description<?> description) {
				return Sets.difference(key.bindings(), description.bindings()).size() == 0;
			}
			
		});
		
		if (descriptions.isEmpty()) {
			logger.warn("Component is missing: Component key [" + key 
					+ "]; You must specify binding annotations on the field correctly");
			return null;
		} else if (descriptions.size() == 1) {
			return component(component(descriptions.toArray(new Description<?>[1])[0]));
		} else {
			descriptions = Collections2.filter(descriptions, new Predicate<Description<?>>() {

				public boolean apply(Description<?> description) {
					for (Annotation binding : description.bindings()) {
						if (binding.annotationType().equals(Default.class)) {
							return true;
						}
					}
					
					return false;
				}
				
			});
			
			if (descriptions.size() == 1) {
				Description<?> description = descriptions.toArray(new Description<?>[1])[0];
				logger.debug("@Default qualified component [" + description + "] is returned in preference");
				return component(component(description));
			} else {
				logger.warn("Component is duplicated: Component key [" + key
						+ "]; You must specify more concrete type " 
						+ "or binding annotations strictly");
				return null;
			}
		}
	}
	
	/**
	 * Returns assembled component instance corresponding to the specified 
	 * description.
	 * <b>NOTE: This method is designed for an internal use.</b>
	 * 
	 * @param <T> The component type.
	 * @param description The {@link Description}.
	 * @return The assembled component instance corresponding to the specified 
	 * description.
	 */
	public <T> T component(Description<T> description) {
		logger.debug("Component is requested: Description [" + description + "]");
		Function<Description<?>, Context> context = new Function<Description<?>, Context>() {

			public Context apply(Description<?> description) {
				Annotation scope = description.scope();
				Class<? extends Annotation> s = (scope == null) ? Prototype.class : scope.annotationType();
				
				if (s.equals(Singleton.class)) {
					return singleton;
				}
				
				Description<?> desc = contexts.get(s);
				
				if (desc == null) {
					Class<? extends Context> ctx = s.getAnnotation(Scope.class).value();
					install(ctx);
					desc = components.get(ctx).get(0);
					contexts.put(s, desc);
				}
				
				return component(component(desc));
			}
			
		};
		Context c = context.apply(description);
		T component = c.get(description);
		
		if (component == null) {
			component = new Assembler<T>(this, description).assemble();
			
			if (component != null) {
				c.put(description, component);
				new Event<Activated>() {}.on(component).fire();
			}
		}
		
		return component;
	}
	
	@SuppressWarnings("unchecked")
	private <T> T component(Object component) {
		return (component instanceof Component) ? ((Component<T>) component).instance() : (T) component;
	}
	
	/**
	 * Assembles the specified component instance by injecting the dependent 
	 * components into it.
	 * The assembled component instance is <b>not</b> managed by the container.
	 * 
	 * @param <T> The component type.
	 * @param component The component instance that the dependent components to 
	 * be injected.
	 * @return The assembled component instance.
	 */
	@SuppressWarnings("unchecked")
	public <T> T assemble(T component) {
		Preconditions.checkArgument(component != null, 
				"Parameter 'component' must not be [" + component + "]");
		Description<T> description = new Description<T>((Class<T>) component.getClass());
		return new Assembler<T>(this, description, component).assemble();
	}
	
	/**
	 * Installs the specified component into the container.
	 * The component must not be an interface and abstract class.
	 * 
	 * @param component The component instance of the specified component class.
	 */
	@SuppressWarnings("unchecked")
	public <T> void install(Class<T> component) {
		Preconditions.checkArgument(component != null, 
				"Parameter 'component' must not be [" + component + "]");
		
		if (installed(component)) {
			logger.warn("Installation ignored: Component [" + component + "] has already been installed");
			return;
		}
		
		int modifiers = component.getModifiers();
		
		if (component.isInterface() || Modifier.isAbstract(modifiers) || !Modifier.isPublic(modifiers)) {
			logger.warn("Interface, abstract and non-public class cannot be installed: " 
					+ "Component [" + component + "]");
			return;
		}
		
		try {
			Description<T> description = new Description<T>(component);
			Set<Class<? extends Annotation>> deployments = description.deployments();
			
			if (!deployments.isEmpty() && !deployments.contains(deployment())) {
				logger.info("Component [" + description
						+ "] is not deployed on ["
						+ deployment().getSimpleName() + "] deployment");
				return;
			}
			
			List<Class<?>> classes = Collections.emptyList();
			
			if (Component.class.isAssignableFrom(component)) {
				Type type = component.getGenericSuperclass();
				
				for (Class<?> superclass = component.getSuperclass(); 
						!superclass.equals(Component.class); superclass = superclass.getSuperclass()) {
					if (type instanceof ParameterizedType) {
						type = ((Class<?>) ((ParameterizedType) type).getRawType()).getGenericSuperclass();
					} else {
						type = ((Class<?>) type).getGenericSuperclass();
					}
				}
				
				Type t = ((ParameterizedType) type).getActualTypeArguments()[0];
				Class<?> clazz = (Class<?>) ((t instanceof ParameterizedType) ? ((ParameterizedType) t).getRawType() : t);
				classes = (List<Class<?>>) ClassUtils.getAllInterfaces(clazz);
				classes.addAll(ClassUtils.getAllSuperclasses(clazz));
				classes.add(clazz);
			} else {
				classes = (List<Class<?>>) ClassUtils.getAllInterfaces(component);
				classes.addAll(ClassUtils.getAllSuperclasses(component));
				classes.add(component);
			}
			
			classes.remove(Object.class);
			
			for (Class<?> type : classes) {
				components.put(type, description);
			}
			
			if (description.interceptor()) {
				for (Annotation intercept : description.intercepts()) {
					interceptors.put(intercept, description);
				}
			}
			
			Annotation scope = description.scope();
			Class<? extends Annotation> s = (scope == null) ? Prototype.class : scope.annotationType();
			
			if (s.equals(Singleton.class) && ((Singleton) description.scope()).eager()) {
				component(description);
			}
			
			installed.add(component);
			logger.info("Component [" + description + "] has been installed");
		} catch (Exception e) {
			logger.warn("Component [" + component + "] cannot be installed", e);
			return;
		}
	}
	
	/**
	 * Indicates the specified component installed or not in the container.
	 * 
	 * @param component The component class to be examined.
	 * @return <code>true</code> if the specified component installed in the 
	 * container.
	 */
	public boolean installed(Class<?> component) {
		Preconditions.checkArgument(component != null, 
				"Parameter 'component' must not be [" + component + "]");
		return installed.contains(component);
	}
	
	/**
	 * Disposes container-managed objects.
	 * This method is invoked by {@link Jazzmaster} in shutdown.
	 */
	public void dispose() {
		singleton.clear();
		components.clear();
		interceptors.clear();
		contexts.clear();
		descriptions.clear();
	}
	
	/**
	 * Returns scope annotation-{@link Context} map the container manages.
	 * 
	 * @return The scope annotation-{@link Context} map the container manages.
	 */
	public Map<Class<? extends Annotation>, Context> contexts() {
		Function<Description<?>, Context> context = new Function<Description<?>, Context>() {

			public Context apply(Description<?> description) {
				Annotation scope = description.scope();
				Class<? extends Annotation> s = (scope == null) ? Prototype.class : scope.annotationType();
				
				if (s.equals(Singleton.class)) {
					return (Context) singleton.get(description);
				} else {
					Context ctx = apply(contexts.get(s));
					return (Context) ctx.get(description);
				}
			}
			
		};
		Map<Class<? extends Annotation>, Context> ctxs = new HashMap<Class<? extends Annotation>, Context>();
		
		for (Class<? extends Annotation> scope : contexts.keySet()) {
			ctxs.put(scope, context.apply(contexts.get(scope)));
		}
		
		return ctxs;
	}
	
	/**
	 * Returns component instance-{@link Description} map the container manages.
	 * 
	 * @return The component instance-{@link Description} map the container 
	 * manages.
	 */
	public WeakHashMap<Object, Description<?>> descriptions() {
		return descriptions;
	}
	
	/**
	 * Returns component type-{@link Description} map the container manages.
	 * 
	 * @return The component type-{@link Description} map the container manages.
	 */
	public Multimap<Class<?>, Description<?>> components() {
		return components;
	}
	
	/**
	 * Returns intercept annotation-{@link Description} map the container 
	 * manages.
	 * 
	 * @return The intercept annotation-{@link Description} map the container 
	 * manages.
	 */
	public ListMultimap<Annotation, Description<?>> interceptors() {
		return interceptors;
	}
	
}
