/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.scope;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.ServletContext;

import org.eiichiro.jazzmaster.Builtin;
import org.eiichiro.jazzmaster.Description;
import org.eiichiro.jazzmaster.inject.Name;
import org.eiichiro.jazzmaster.lifecycle.Destroyed;
import org.eiichiro.jazzmaster.lifecycle.Event;
import org.eiichiro.jazzmaster.lifecycle.Passivated;

/**
 * Built-in {@link Context} implementation corresponding to the 
 * {@link Application} scope.
 * {@code ApplicationContext} stores a component in servlet context. If the 
 * application is not running in Web environment, this context becomes 
 * equal to {@link SingletonContext}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Builtin
@Singleton(eager=true)
@Name("org.eiichiro.jazzmaster.scope.ApplicationContext")
public class ApplicationContext extends SingletonContext implements WebContext<ServletContext> {

	private ServletContext context;
	
	/**
	 * Returns the component corresponding to the specified {@link Description} 
	 * from the current {@code ServletContext}.
	 * If the application is not running in Web environment, the component is 
	 * got from {@code SingletonContext}'s component store.
	 * 
	 * @param <T> The component type.
	 * @param description The description corresponding to the component.
	 * @return The component corresponding to the specified {@link Description}.
	 */
	@SuppressWarnings("unchecked")
	public <T> T get(Description<T> description) {
		if (context == null) {
			return super.get(description);
		}
		
		Map<Description<?>, Object> components 
				= (Map<Description<?>, Object>) context.getAttribute(COMPONENTS);
		return (T) components.get(description);
	}

	/**
	 * Puts the specified component and description pair into the current 
	 * {@code ServletContext}.
	 * If the application is not running in Web environment, the component is 
	 * put into {@code SingletonContext}'s component store.
	 * 
	 * @param <T> The component type.
	 * @param description The description corresponding to the component.
	 * @param component The component to be put.
	 */
	@SuppressWarnings("unchecked")
	public <T> void put(Description<T> description, T component) {
		if (context == null) {
			super.put(description, component);
			return;
		}
		
		Map<Description<?>, Object> components 
				= (Map<Description<?>, Object>) context.getAttribute(COMPONENTS);
		components.put(description, component);
	}
	
	/**
	 * Associates the current {@code ServletContext} with this instance.
	 * 
	 * @param context The current {@code ServletContext}.
	 */
	public void associate(ServletContext context) {
		context.setAttribute(COMPONENTS, new ConcurrentHashMap<Description<?>, Object>());
		this.context = context;
	}

	/**
	 * Deassociates the current {@code ServletContext} from this instance.
	 * 
	 * @param context The current {@code ServletContext}.
	 */
	@SuppressWarnings("unchecked")
	public void deassociate(ServletContext context) {
		Map<Description<?>, Object> components 
				= (Map<Description<?>, Object>) context.getAttribute(COMPONENTS);
		new Event<Passivated>() {}.on(components.values()).fire();
		context.removeAttribute(COMPONENTS);
		new Event<Destroyed>() {}.on(components.values()).fire();
		this.context = null;
	}

}
