/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.scope;

import java.util.Collection;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.eiichiro.jazzmaster.Builtin;
import org.eiichiro.jazzmaster.Description;
import org.eiichiro.jazzmaster.inject.Name;
import org.eiichiro.jazzmaster.lifecycle.Destroyed;
import org.eiichiro.jazzmaster.lifecycle.Event;
import org.eiichiro.jazzmaster.lifecycle.Passivated;

/**
 * Built-in {@link Context} implementation corresponding to the 
 * {@link Singleton} scope.
 * {@code SingletonContext} stores a component in static global store. So the 
 * components stored in this context are sustained in per-{@code ClassLoader}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Builtin
@Singleton
@Name("org.eiichiro.jazzmaster.scope.SingletonContext")
public class SingletonContext implements Context {

	private static Map<Description<?>, Object> store 
			= new ConcurrentHashMap<Description<?>, Object>();
	
	/**
	 * Returns the component corresponding to the specified {@link Description} 
	 * from the static global store.
	 * 
	 * @param <T> The component type.
	 * @param description The description corresponding to the component.
	 * @return The component corresponding to the specified {@link Description}.
	 */
	@SuppressWarnings("unchecked")
	public <T> T get(Description<T> description) {
		return (T) store.get(description);
	}

	/**
	 * Puts the specified component and description pair into the static global 
	 * store.
	 * 
	 * @param <T> The component type.
	 * @param description The description corresponding to the component.
	 * @param component The component to be put.
	 */
	public <T> void put(Description<T> description, T component) {
		store.put(description, component);
	}
	
	/**
	 * Clears component store.
	 * Raises {@link Passivated} event then {@link Destroyed} event on the 
	 * components.
	 */
	public void clear() {
		Collection<Object> components = store.values();
		new Event<Passivated>() {}.on(components).fire();
		new Event<Destroyed>() {}.on(components).fire();
		store.clear();
	}

}
